@@ -35,8 +35,6 @@
 #include "choose.h"
 #include "gdm-net.h"
 
-static const gchar RCSid[]="$Id: display.c,v 1.23 2001/10/26 03:43:28 jirka Exp $";
-
 /* External vars */
 extern gboolean GdmXdmcp;
 extern gint sessions;
@@ -193,7 +191,8 @@ gdm_display_manage (GdmDisplay *d)
 	    gdm_server_stop (d);
 	    /* we expect to return after the session finishes */
 	    _exit (DISPLAY_REMANAGE);
-	} else if (d->type == TYPE_XDMCP && d->dispstat == XDMCP_MANAGED) {
+	} else if (d->type == TYPE_XDMCP &&
+		   d->dispstat == XDMCP_MANAGED) {
 	    gdm_slave_start (d);
 	    gdm_server_stop (d);
 	    /* we expect to return after the session finishes */
@@ -219,7 +218,7 @@ gdm_display_manage (GdmDisplay *d)
     }
 
     if (SERVER_IS_LOCAL (d)) {
-	    d->dispstat = DISPLAY_SUCCESS;
+	    d->dispstat = DISPLAY_ALIVE;
     }
 
     /* reset sleep to 0 */
