@@ -44,9 +44,12 @@
 #include <sys/ioctl.h>
 #include <net/if.h>
 
+#include <viciousui.h>
+
 #include "gdmchooser.h"
 #include "gdm.h"
 #include "misc.h"
+#include "gdmwm.h"
 
 /* set the DOING_GDM_DEVELOPMENT env variable if you want to
  * search for the glade file in the current dir and not the system
@@ -54,22 +57,35 @@
  * in the source and recompile */
 static gboolean DOING_GDM_DEVELOPMENT = FALSE;
 
-
-static const gchar RCSid[]="$Id: gdmchooser.c,v 1.15 2001/05/21 18:30:37 jirka Exp $";
+static gboolean RUNNING_UNDER_GDM = FALSE;
 
 static const gchar *scanning_message = N_("Please wait: scanning local network for XDMCP-enabled hosts...");
 static const gchar *empty_network = N_("No serving hosts were found.");
 static const gchar *active_network = N_("Choose a host to connect to from the selection below.");
 static gchar *glade_filename = NULL;
 
-static GdmChooserHost *gdm_chooser_host_alloc (gchar *hostname, gchar *description);
-static void gdm_chooser_decode_packet (void);
-static void gdm_chooser_abort (const gchar *format, ...);
+static GdmChooserHost * gdm_chooser_host_alloc (const char *hostname,
+						const char *description,
+						struct in_addr *ia,
+						gboolean willing);
+
+static gboolean gdm_chooser_decode_packet (GIOChannel   *source,
+					   GIOCondition  condition,
+					   gpointer      data);
+
+static void gdm_chooser_abort (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 static void gdm_chooser_browser_update (void);
-static void gdm_chooser_xdmcp_init (gchar **hosts);
+static void gdm_chooser_xdmcp_init (char **hosts);
 static void gdm_chooser_host_dispose (GdmChooserHost *host);
-static void gdm_chooser_choose_host (gchar *hostname);
-static void gdm_chooser_add_hosts (gchar **hosts);
+static void gdm_chooser_choose_host (const gchar *hostname);
+static void gdm_chooser_add_hosts (char **hosts);
+
+static guint scan_time_handler = 0;
+
+#define PING_TIMEOUT 2000
+#define PING_TRIES 3
+static int ping_tries = PING_TRIES;
+static guint ping_try_handler = 0;
 
 /* Fixetyfix */
 int XdmcpReallocARRAY8 (ARRAY8Ptr array, int length);
@@ -92,6 +108,7 @@ static XdmcpBuffer querybuf;
 static GSList *bcaddr;
 static GSList *queryaddr;
 
+static gint  GdmXineramaScreen;
 static gint  GdmIconMaxHeight;
 static gint  GdmIconMaxWidth;
 static gboolean  GdmDebug;
@@ -99,19 +116,26 @@ static gint  GdmScanTime;
 static gchar *GdmHostIconDir;
 static gchar *GdmHostDefaultIcon;
 static gchar *GdmGtkRC;
+static gchar *GdmHosts;
+static gboolean GdmBroadcast;
+static gchar *GdmBackgroundColor;
+static int GdmBackgroundType;
+enum {
+	GDM_BACKGROUND_NONE = 0,
+	GDM_BACKGROUND_IMAGE = 1,
+	GDM_BACKGROUND_COLOR = 2
+};
 
 static GladeXML *chooser_app;
 static GtkWidget *chooser, *manage, *rescan, *cancel;
 static GtkWidget *status_label;
 
-static gint maxwidth = 0;
 static GIOChannel *channel;
 static GList *hosts = NULL;
 static GdkImlibImage *defhostimg;
 static GnomeIconList *browser;
 static GdmChooserHost *curhost;
 
-
 static gint 
 gdm_chooser_sort_func (gpointer d1, gpointer d2)
 {
@@ -124,9 +148,37 @@ gdm_chooser_sort_func (gpointer d1, gpointer d2)
     return strcmp (a->name, b->name);
 }
 
+static gboolean
+gdm_host_known (struct in_addr *ia)
+{
+	GList *li;
 
-static void
-gdm_chooser_decode_packet (void)
+	for (li = hosts; li != NULL; li = li->next) {
+		GdmChooserHost *host = li->data;
+		if (memcmp (&host->ia, ia, sizeof (struct in_addr)) == 0) {
+			return TRUE;
+		}
+	}
+	return FALSE;
+}
+
+static gboolean
+is_loopback_addr (struct in_addr *ia)
+{
+	const char lo[] = {127,0,0,1};
+
+	if (ia->s_addr == INADDR_LOOPBACK ||
+	    memcmp (&ia->s_addr, lo, 4) == 0) {
+		return TRUE;
+	} else {
+		return FALSE;
+	}
+}
+
+static gboolean
+gdm_chooser_decode_packet (GIOChannel   *source,
+			   GIOCondition  condition,
+			   gpointer      data)
 {
     struct sockaddr_in clnt_sa;
     gint sa_len = sizeof (clnt_sa);
@@ -136,63 +188,73 @@ gdm_chooser_decode_packet (void)
     gchar *hostname = NULL;
     gchar *status = NULL;
     ARRAY8 auth, host, stat;
-    GdmChooserHost *new_host;
 
     if (! XdmcpFill (sockfd, &buf, (XdmcpNetaddr) &clnt_sa, &sa_len))
-	return;
+	return TRUE;
 
     if (! XdmcpReadHeader (&buf, &header))
-	return;
+	return TRUE;
     
     if (header.version != XDM_PROTOCOL_VERSION)
-	return;
+	return TRUE;
 
-    if (header.opcode != WILLING)
-	return;
-    
-    if (! XdmcpReadARRAY8 (&buf, &auth))
-	goto done;
-    
-    if (! XdmcpReadARRAY8 (&buf, &host))
-	goto done;
-    
-    if (! XdmcpReadARRAY8 (&buf, &stat))
-	goto done;
-    
-    status = g_strndup (stat.data, stat.length);
-    
-    he = gethostbyaddr ((gchar *) &clnt_sa.sin_addr, 
-			sizeof (struct in_addr), 
-			AF_INET);
-    
-    hostname = (he && he->h_name) ? he->h_name : inet_ntoa (clnt_sa.sin_addr);
-    
-    /* We can't pipe hostnames larger than this */
-    if (strlen (hostname)+1 > PIPE_BUF)
-	goto done;
+    if (header.opcode == WILLING) {
+	    if (! XdmcpReadARRAY8 (&buf, &auth))
+		    goto done;
 
-    new_host = gdm_chooser_host_alloc (hostname, (gchar *) status);
-    if (g_list_find_custom (hosts,
-			    new_host,
-			    (GCompareFunc) gdm_chooser_sort_func)) {
-	    gdm_chooser_host_dispose (new_host);
-	    goto done;
+	    if (! XdmcpReadARRAY8 (&buf, &host))
+		    goto done;
+
+	    if (! XdmcpReadARRAY8 (&buf, &stat))
+		    goto done;
+
+	    status = g_strndup (stat.data, stat.length);
+    } else if (header.opcode == UNWILLING) {
+	    /* immaterial, will not be shown */
+	    status = NULL;
+    } else {
+	    return TRUE;
     }
-    
-    hosts = g_list_insert_sorted (hosts, 
-				  new_host,
-				  (GCompareFunc) gdm_chooser_sort_func);
+
+    if ( ! is_loopback_addr (&clnt_sa.sin_addr)) {
+	    he = gethostbyaddr ((gchar *) &clnt_sa.sin_addr,
+				sizeof (struct in_addr),
+				AF_INET);
+
+	    hostname = (he && he->h_name) ? he->h_name : inet_ntoa (clnt_sa.sin_addr);
+
+	    /* We can't pipe hostnames larger than this */
+	    if (strlen (hostname)+1 > PIPE_BUF)
+		    goto done;
+
+	    hostname = g_strdup (hostname);
+    } else {
+	    hostname = g_new0 (char, 1024);
+	    if (gethostname (hostname, 1023) != 0) {
+		    g_free (hostname);
+		    goto done;
+	    }
+    }
+
+    gdm_chooser_host_alloc (hostname,
+			    status,
+			    &clnt_sa.sin_addr,
+			    header.opcode == WILLING);
+
+    g_free (hostname);
 
     gdm_chooser_browser_update ();
     
  done:
-    XdmcpDisposeARRAY8 (&auth);
-    XdmcpDisposeARRAY8 (&host);
-    XdmcpDisposeARRAY8 (&stat);
+    if (header.opcode == WILLING) {
+	    XdmcpDisposeARRAY8 (&auth);
+	    XdmcpDisposeARRAY8 (&host);
+	    XdmcpDisposeARRAY8 (&stat);
+    }
     
     g_free (status);
     
-    return;
+    return TRUE;
 }
 
 
@@ -247,21 +309,65 @@ gdm_chooser_find_bcaddr (void)
 	}
 }
 
+static gboolean
+chooser_scan_time_update (gpointer data)
+{
+	scan_time_handler = 0;
+	gdm_chooser_browser_update ();
+	return FALSE;
+}
 
-gboolean 
-gdm_chooser_xdmcp_discover (void)
+static void
+do_ping (gboolean full)
 {
     struct sockaddr_in sock;
     GSList *bl = bcaddr;
     GSList *ql = queryaddr;
     struct in_addr *ia;
+
+    sock.sin_family = AF_INET;
+    sock.sin_port = htons (XDM_UDP_PORT);
+
+    while (bl) {
+	    ia = (struct in_addr *) bl->data;
+	    sock.sin_addr.s_addr = ia->s_addr; 
+	    XdmcpFlush (sockfd, &bcbuf, (XdmcpNetaddr) &sock, (int)sizeof (struct sockaddr_in));
+	    bl = bl->next;
+    }
+
+    while (ql != NULL) {
+	    ia = (struct in_addr *) ql->data;
+	    if (full ||
+		! gdm_host_known (ia)) {
+		    sock.sin_addr.s_addr = ia->s_addr;
+		    XdmcpFlush (sockfd, &querybuf, (XdmcpNetaddr) &sock, (int)sizeof (struct sockaddr_in));
+	    }
+	    ql = ql->next;
+    }
+}
+
+static gboolean
+ping_try (gpointer data)
+{
+	do_ping (FALSE);
+
+	ping_tries --;
+	if (ping_tries <= 0)
+		return FALSE;
+	else
+		return TRUE;
+}
+
+gboolean 
+gdm_chooser_xdmcp_discover (void)
+{
     GList *hl = hosts;
 
     gtk_widget_set_sensitive (GTK_WIDGET (manage), FALSE);
     gnome_icon_list_clear (GNOME_ICON_LIST (browser));
     gtk_widget_set_sensitive (GTK_WIDGET (browser), FALSE);
     gtk_label_set (GTK_LABEL (status_label),
-		   scanning_message);
+		   _(scanning_message));
 
     while (hl) {
 	gdm_chooser_host_dispose ((GdmChooserHost *) hl->data);
@@ -272,26 +378,18 @@ gdm_chooser_xdmcp_discover (void)
 
     hosts = NULL;
 
-    sock.sin_family = AF_INET;
-    sock.sin_port = htons (XDM_UDP_PORT);
+    do_ping (TRUE);
 
-    while (bl) {
-	ia = (struct in_addr *) bl->data;
-	sock.sin_addr.s_addr = ia->s_addr; 
-	XdmcpFlush (sockfd, &bcbuf, (XdmcpNetaddr) &sock, sizeof (struct sockaddr_in));
-	bl = bl->next;
-    }
+    if (scan_time_handler > 0)
+	    g_source_remove (scan_time_handler);
+    scan_time_handler = g_timeout_add (GdmScanTime * 1000, 
+				       chooser_scan_time_update, NULL);
 
-    /*
-    tid = g_timeout_add (GdmScanTime * 1000, 
-			 (GSourceFunc) gdm_chooser_browser_update, NULL);
-     */
-    while (ql != NULL) {
-	    ia = (struct in_addr *) ql->data;
-	    sock.sin_addr.s_addr = ia->s_addr;
-	    XdmcpFlush (sockfd, &querybuf, (XdmcpNetaddr) &sock, sizeof (struct sockaddr_in));
-	    ql = ql->next;
-    }
+    /* Note we already used up one try */
+    ping_tries = PING_TRIES - 1;
+    if (ping_try_handler > 0)
+	    g_source_remove (ping_try_handler);
+    ping_try_handler = g_timeout_add (PING_TIMEOUT, ping_try, NULL);
 
     return TRUE;
 }
@@ -329,7 +427,7 @@ gdm_chooser_xdmcp_init (char **hosts)
     channel = g_io_channel_unix_new (sockfd);
     g_io_add_watch_full (channel, G_PRIORITY_DEFAULT,
 			G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
-			(GIOFunc) gdm_chooser_decode_packet,
+			gdm_chooser_decode_packet,
 			GINT_TO_POINTER (sockfd), NULL);
     g_io_channel_unref (channel);
 
@@ -350,6 +448,11 @@ gdm_chooser_cancel (void)
 gboolean
 gdm_chooser_manage (GtkButton *button, gpointer data)
 {
+    if (scan_time_handler > 0) {
+	    g_source_remove (scan_time_handler);
+	    scan_time_handler = 0;
+    }
+
     if (curhost)
       gdm_chooser_choose_host (curhost->name);
    
@@ -370,7 +473,7 @@ gdm_chooser_abort (const gchar *format, ...)
     s = g_strdup_vprintf (format, args);
     va_end (args);
 
-    syslog (LOG_ERR, s);
+    syslog (LOG_ERR, "%s", s);
     closelog ();
 
     exit (EXIT_FAILURE);
@@ -387,6 +490,7 @@ gdm_chooser_parse_config (void)
 
     gnome_config_push_prefix ("=" GDM_CONFIG_FILE "=/");
 
+    GdmXineramaScreen = gnome_config_get_int (GDM_KEY_XINERAMASCREEN);
     GdmGtkRC = gnome_config_get_string (GDM_KEY_GTKRC);
     GdmScanTime = gnome_config_get_int (GDM_KEY_SCAN);
     GdmHostDefaultIcon = gnome_config_get_string (GDM_KEY_HOST);
@@ -395,6 +499,24 @@ gdm_chooser_parse_config (void)
     GdmIconMaxHeight = gnome_config_get_int (GDM_KEY_ICONHEIGHT);
     GdmDebug = gnome_config_get_bool (GDM_KEY_DEBUG);
 
+    GdmBackgroundColor = gnome_config_get_string (GDM_KEY_BACKGROUNDCOLOR);
+    GdmBackgroundType = gnome_config_get_int (GDM_KEY_BACKGROUNDTYPE);
+
+    /* note that command line arguments will prevail over these */
+    GdmHosts = gnome_config_get_string (GDM_KEY_HOSTS);
+    GdmBroadcast = gnome_config_get_bool (GDM_KEY_BROADCAST);
+    /* if broadcasting, then append BROADCAST to hosts */
+    if (GdmBroadcast) {
+	    if (ve_string_empty (GdmHosts)) {
+		    g_free (GdmHosts);
+		    GdmHosts = "BROADCAST";
+	    } else {
+		    char *tmp = g_strconcat (GdmHosts, ",BROADCAST", NULL);
+		    g_free (GdmHosts);
+		    GdmHosts = tmp;
+	    }
+    }
+
     if (GdmScanTime < 1) GdmScanTime = 1;
     if (GdmIconMaxWidth < 0) GdmIconMaxWidth = 128;
     if (GdmIconMaxHeight < 0) GdmIconMaxHeight = 128;
@@ -402,6 +524,24 @@ gdm_chooser_parse_config (void)
     gnome_config_pop_prefix();
 }
 
+static GdmChooserHost *
+gdm_nth_willing_host (int n)
+{
+	GList *li;
+	int i;
+
+	i = 0;
+	for (li = hosts; li != NULL; li = li->next) {
+		GdmChooserHost *host = li->data;
+		if (host->willing) {
+			if (i == n)
+				return host;
+			i++;
+		}
+	}
+	return NULL;
+}
+
 
 gboolean 
 gdm_chooser_browser_select (GtkWidget *widget, gint selected, GdkEvent *event)
@@ -413,9 +553,14 @@ gdm_chooser_browser_select (GtkWidget *widget, gint selected, GdkEvent *event)
 	
     case GDK_BUTTON_PRESS:
     case GDK_BUTTON_RELEASE:
-	curhost = g_list_nth_data (hosts, selected);
+	curhost = gdm_nth_willing_host (selected);
 	gtk_widget_set_sensitive (manage, TRUE);
 	break;
+
+    case GDK_2BUTTON_PRESS:
+	curhost = gdm_nth_willing_host (selected);
+	gdm_chooser_manage (NULL, NULL);
+	break;
 	
     default: 
 	break;
@@ -450,32 +595,36 @@ gdm_chooser_browser_unselect (GtkWidget *widget, gint selected, GdkEvent *event)
 static void
 gdm_chooser_browser_update (void)
 {
-    GList *list = hosts;
+    GList *li;
+    gboolean any;
 
     gnome_icon_list_freeze (GNOME_ICON_LIST (browser));
     gnome_icon_list_clear (GNOME_ICON_LIST (browser));
 
-    while (list) {
-	GdmChooserHost *host;
-	gchar *temp;
+    any = FALSE;
+    for (li = hosts; li != NULL; li = li->next) {
+	    GdmChooserHost *host = (GdmChooserHost *) li->data;
 
-	host = (GdmChooserHost *) list->data;
-
-	temp = g_strconcat (host->name, "\n", host->desc, NULL);
-	gnome_icon_list_append_imlib (GNOME_ICON_LIST (browser), host->picture, temp);
-	g_free (temp);
-
-	list = list->next;
+	    if (host->willing) {
+		    /* FIXME: the \n doesn't actually propagate
+		     * since the icon list is a broken piece of horsedung */
+		    char *temp = g_strconcat (host->name, " \n",
+					      host->desc, NULL);
+		    gnome_icon_list_append_imlib (GNOME_ICON_LIST (browser),
+						  host->picture, temp);
+		    g_free (temp);
+		    any = TRUE;
+	    }
     }
 
     gnome_icon_list_thaw (GNOME_ICON_LIST (browser));
 
-    if (hosts) {
+    if (any) {
       gtk_label_set (GTK_LABEL (glade_xml_get_widget (chooser_app, "status_label")),
-		     active_network);
+		     _(active_network));
     } else {
       gtk_label_set (GTK_LABEL (glade_xml_get_widget (chooser_app, "status_label")),
-		     empty_network);
+		     _(empty_network));
     }
     gtk_widget_set_sensitive (GTK_WIDGET (manage), FALSE);
     gtk_widget_set_sensitive (GTK_WIDGET (rescan), TRUE);
@@ -491,10 +640,8 @@ display_chooser_information (void)
 static void 
 gdm_chooser_gui_init (void)
 {
-    GdkWindow *rootwin;
     GtkStyle  *style;
     GdkColor  bbg = { 0, 0xFFFF, 0xFFFF, 0xFFFF };
-    struct    stat statbuf;
  
    
         /* Look for the glade file in $(datadir)/gdmconfig or, failing that,
@@ -521,15 +668,10 @@ gdm_chooser_gui_init (void)
 	gtk_rc_parse (GdmGtkRC);
 
     /* Load default host image */
-    if (stat (GdmHostDefaultIcon, &statbuf))
+    if (access (GdmHostDefaultIcon, R_OK) != 0)
 	gdm_chooser_abort (_("Can't open default host icon: %s"), GdmHostDefaultIcon);
-    else {
+    else
 	defhostimg = gdk_imlib_load_image (GdmHostDefaultIcon);
-	maxwidth = defhostimg->rgb_width;
-    }
-
-    /* Root Window */
-    rootwin = gdk_window_foreign_new (GDK_ROOT_WINDOW ());
 
     /* Main window */
     g_assert (glade_filename != NULL);
@@ -573,15 +715,11 @@ gdm_chooser_gui_init (void)
     style->bg[GTK_STATE_NORMAL] = bbg;
     gtk_widget_push_style (style);
 
-    /* Icon list */
-    if (maxwidth < GdmIconMaxWidth/2)
-	maxwidth = (gint) GdmIconMaxWidth/2;
-
     browser = (GnomeIconList *) glade_xml_get_widget (chooser_app, 
 						     "chooser_iconlist");
     gnome_icon_list_freeze (GNOME_ICON_LIST (browser));
     gnome_icon_list_set_separators (GNOME_ICON_LIST (browser), " /-_.");
-    gnome_icon_list_set_icon_width (GNOME_ICON_LIST (browser), maxwidth+20);
+    gnome_icon_list_set_icon_width (GNOME_ICON_LIST (browser), GdmIconMaxWidth + 20);
     gnome_icon_list_set_icon_border (GNOME_ICON_LIST (browser), 2);
     gtk_widget_pop_style();
     gnome_icon_list_thaw (GNOME_ICON_LIST (browser));
@@ -590,8 +728,7 @@ gdm_chooser_gui_init (void)
 			  (gint) gdk_screen_width() * 0.4, 
 			  (gint) gdk_screen_height() * 0.6);
 
-    /* FIXME: This should probably obey some user-specified geometry? */
-    gtk_window_position (GTK_WINDOW (chooser), GTK_WIN_POS_CENTER);
+    gdm_wm_center_window (GTK_WINDOW (chooser));
 }
 
 
@@ -624,20 +761,43 @@ gdm_chooser_signals_init (void)
 }
 
 static GdmChooserHost * 
-gdm_chooser_host_alloc (gchar *hostname, gchar *description)
+gdm_chooser_host_alloc (const char *hostname,
+			const char *description,
+			struct in_addr *ia,
+			gboolean willing)
 {
     GdmChooserHost *host;
     GdkImlibImage *imlibimg;
     gchar *hostimg;
-    struct stat statbuf;
+    GList *hostl;
 
     host = g_malloc (sizeof (GdmChooserHost));
     host->name = g_strdup (hostname);
     host->desc = g_strdup (description);
+    memcpy (&host->ia, ia, sizeof (struct in_addr));
+    host->willing = willing;
+
+    hostl = g_list_find_custom (hosts,
+				host,
+				(GCompareFunc) gdm_chooser_sort_func);
+    /* replace */
+    if (hostl != NULL) {
+	    GdmChooserHost *old = hostl->data;
+	    hostl->data = host;
+	    gdm_chooser_host_dispose (old);
+    } else {
+	    hosts = g_list_insert_sorted (hosts, 
+					  host,
+					  (GCompareFunc) gdm_chooser_sort_func);
+    }
+    
+    if ( ! willing)
+	    return host;
 
     hostimg = g_strconcat (GdmHostIconDir, "/", hostname, NULL);
 
-    if (!stat (hostimg, &statbuf) && (imlibimg = gdk_imlib_load_image (hostimg))) {
+    if (access (hostimg, R_OK) == 0 &&
+	(imlibimg = gdk_imlib_load_image (hostimg))) {
 	gint w, h;
 
 	w = imlibimg->rgb_width;
@@ -646,20 +806,17 @@ gdm_chooser_host_alloc (gchar *hostname, gchar *description)
 	if (w>h && w>GdmIconMaxWidth) {
 	    h = h * ((gfloat) GdmIconMaxWidth/w);
 	    w = GdmIconMaxWidth;
-	} 
-	else if (h>GdmIconMaxHeight) {
+	} else if (h>GdmIconMaxHeight) {
 	    w = w * ((gfloat) GdmIconMaxHeight/h);
 	    h = GdmIconMaxHeight;
 	}
 
-	maxwidth = MAX (maxwidth, w);
-
 	host->picture = gdk_imlib_clone_scaled_image (imlibimg, w, h);
 	
 	gdk_imlib_destroy_image (imlibimg);
-    }
-    else
+    } else {
 	host->picture = defhostimg;
+    }
 
     g_free (hostimg);
 
@@ -675,9 +832,12 @@ gdm_chooser_host_dispose (GdmChooserHost *host)
 
     if (host->picture != defhostimg)
 	gdk_imlib_destroy_image (host->picture);
+    host->picture = NULL;
 
     g_free (host->name);
+    host->name = NULL;
     g_free (host->desc);
+    host->desc = NULL;
     g_free (host);
 }
 
@@ -687,9 +847,8 @@ static gint connection_type = 0;
 
 struct poptOption xdm_options [] = {
   { "xdmaddress", '\0', POPT_ARG_STRING|POPT_ARGFLAG_ONEDASH, &xdm_address, 0,
-    "setting socket for xdm communication", "xdm response socket" },
-  { "clientaddress", '\0', POPT_ARG_STRING|POPT_ARGFLAG_ONEDASH, &client_address, 0, "setting client address to return in response to xdm", "client address" },
-  { "connectionType", '\0', POPT_ARG_INT|POPT_ARGFLAG_ONEDASH, &connection_type, 0, "setting connection type to return in response to xdm", "connection type" },
+    N_("Socket for xdm communication"), N_("SOCKET") }, { "clientaddress", '\0', POPT_ARG_STRING|POPT_ARGFLAG_ONEDASH, &client_address, 0, N_("Client address to return in response to xdm"), N_("ADDRESS") },
+  { "connectionType", '\0', POPT_ARG_INT|POPT_ARGFLAG_ONEDASH, &connection_type, 0, N_("Connection type to return in response to xdm"), N_("TYPE") },
   POPT_AUTOHELP
   { NULL, 0, 0, NULL, 0}
 };
@@ -700,7 +859,7 @@ struct poptOption xdm_options [] = {
 #define HexChar(c)  ('0' <= (c) && (c) <= '9' ? (c) - '0' : (c) - 'a' + 10)
 
 static int
-from_hex (gchar *s, gchar *d, int len)
+from_hex (const char *s, char *d, int len)
 {
   int t;
   while (len >= 2)
@@ -721,7 +880,7 @@ from_hex (gchar *s, gchar *d, int len)
 
 
 static void
-gdm_chooser_choose_host (gchar *hostname)
+gdm_chooser_choose_host (const char *hostname)
 {
   ARRAY8 tmparr;
   struct hostent *hentry;
@@ -736,8 +895,9 @@ gdm_chooser_choose_host (gchar *hostname)
       char buf[1024];
       XdmcpBuffer buffer;
       long family, port, addr;
-      if (strlen (xdm_address) > 64 || from_hex (xdm_address, xdm_addr, strlen (xdm_address)) != 0)
- gdm_chooser_abort ("gdm_chooser_chooser_host: Invalid xdm address.");
+      if (strlen (xdm_address) > 64 ||
+	  from_hex (xdm_address, xdm_addr, strlen (xdm_address)) != 0)
+	      gdm_chooser_abort ("gdm_chooser_chooser_host: Invalid xdm address.");
 
       family = (xdm_addr[0] << 8) | xdm_addr[1];
       port = (xdm_addr[2] << 8) | xdm_addr[3];
@@ -746,24 +906,24 @@ gdm_chooser_choose_host (gchar *hostname)
       in_addr.sin_port = htons (port);
       in_addr.sin_addr.s_addr = htonl (addr);
       if ((fd = socket (PF_INET, SOCK_STREAM, 0)) == -1)
- gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't create response socket.");
+	      gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't create response socket.");
 
-      if (connect (fd, (struct sockaddr_in *) &in_addr, sizeof (in_addr)) == -1)
- gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't connect to xdm.");
+      if (connect (fd, (struct sockaddr *) &in_addr, sizeof (in_addr)) == -1)
+	      gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't connect to xdm.");
       buffer.data = (BYTE *) buf;
       buffer.size = sizeof (buf);
       buffer.pointer = 0;
       buffer.count = 0;
 
       if (strlen (client_address) > 64 || from_hex (client_address, client_addr, strlen (client_address)) != 0)
- gdm_chooser_abort ("gdm_chooser_chooser_host: Invalid client address.");
+	      gdm_chooser_abort ("gdm_chooser_chooser_host: Invalid client address.");
       tmparr.data = (BYTE *) client_addr;
       tmparr.length = strlen (client_address) / 2;
       XdmcpWriteARRAY8 (&buffer, &tmparr);
       XdmcpWriteCARD16 (&buffer, (CARD16) connection_type);
       hentry = gethostbyname (hostname);
       if (!hentry)
- gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't get host entry for %s", hostname);
+	      gdm_chooser_abort ("gdm_chooser_chooser_host: Couldn't get host entry for %s", hostname);
 
       tmparr.data = (BYTE *) hentry->h_addr_list[0]; /* XXX */
       tmparr.length = 4;
@@ -775,7 +935,7 @@ gdm_chooser_choose_host (gchar *hostname)
 }
 
 static void
-gdm_chooser_add_hosts (gchar **hosts)
+gdm_chooser_add_hosts (char **hosts)
 {
 	struct hostent *hostent;
 	struct sockaddr_in qa;
@@ -783,7 +943,7 @@ gdm_chooser_add_hosts (gchar **hosts)
 	int i;
 
 	for (i = 0; hosts != NULL && hosts[i] != NULL; i++) {
-		char *name = hosts[i];
+		const char *name = hosts[i];
 
 		if (strcmp (name, "BROADCAST") == 0) {
 			gdm_chooser_find_bcaddr ();
@@ -813,16 +973,46 @@ gdm_chooser_add_hosts (gchar **hosts)
 		gdm_chooser_find_bcaddr ();
 }
 
+static void
+set_background (void)
+{
+	if (GdmBackgroundType != GDM_BACKGROUND_NONE) {
+		GdkColor color;
+		GdkColormap *colormap;
+
+		if (ve_string_empty (GdmBackgroundColor) ||
+		    ! gdk_color_parse (GdmBackgroundColor, &color)) {
+			gdk_color_parse ("#007777", &color);
+		}
+
+		colormap = gdk_window_get_colormap (GDK_ROOT_PARENT ());
+		/* paranoia */
+		if (colormap != NULL) {
+			gdk_error_trap_push ();
+
+			gdk_color_alloc (colormap, &color);
+			gdk_window_set_background (GDK_ROOT_PARENT (), &color);
+			gdk_window_clear (GDK_ROOT_PARENT ());
+
+			gdk_flush ();
+			gdk_error_trap_pop ();
+		}
+	}
+}
+
 int 
 main (int argc, char *argv[])
 {
-    gchar **fixedargv;
-    gint fixedargc, i;
-    gchar **hosts;
+    char **fixedargv;
+    int fixedargc, i;
+    char **hosts;
     poptContext ctx;
+    const char *gdm_version;
 
     if (g_getenv ("DOING_GDM_DEVELOPMENT") != NULL)
 	    DOING_GDM_DEVELOPMENT = TRUE;
+    if (g_getenv ("RUNNING_UNDER_GDM") != NULL)
+	    RUNNING_UNDER_GDM = TRUE;
 
     /* Avoid creating ~gdm/.gnome stuff */
     gnome_do_not_create_directories = TRUE;
@@ -843,16 +1033,66 @@ main (int argc, char *argv[])
     bindtextdomain (PACKAGE, GNOMELOCALEDIR);
     textdomain (PACKAGE);
 
-    gnome_preferences_set_dialog_position(GTK_WIN_POS_CENTER);
+    gdm_wm_screen_init (GdmXineramaScreen);
+
+    gdm_version = g_getenv ("GDM_VERSION");
+
+    if (gdm_version != NULL &&
+	strcmp (gdm_version, VERSION) != 0) {
+	    char *msg;
+	    GtkWidget *dialog;
+
+	    gdm_wm_init (0);
+
+	    gdm_wm_focus_new_windows (TRUE);
+
+	    msg = g_strdup_printf
+		    (_("The chooser version (%s) does not match the daemon "
+		       "version (%s).\n"
+		       "You have probably just upgraded gdm.\n"
+		       "Please restart the gdm daemon or reboot the computer."),
+		     VERSION, gdm_version);
+	    dialog = gnome_error_dialog (msg);
+	    g_free (msg);
+
+	    gtk_widget_show_all (dialog);
+	    gdm_wm_center_window (GTK_WINDOW (dialog));
+	    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
+
+	    return EXIT_SUCCESS;
+    }
     
     gdm_chooser_parse_config();
     gdm_chooser_gui_init();
     gdm_chooser_signals_init();
 
-    hosts = (gchar **) poptGetArgs (ctx);
+    set_background ();
+
+    hosts = (char **)poptGetArgs (ctx);
+    /* when no hosts on the command line, take them from the config */
+    if (hosts == NULL ||
+	hosts[0] == NULL) {
+	    int i;
+	    hosts = g_strsplit (GdmHosts, ",", -1);
+	    for (i = 0; hosts != NULL && hosts[i] != NULL; i++) {
+		    g_strstrip (hosts[i]);
+	    }
+    }
     gdm_chooser_xdmcp_init (hosts);
     poptFreeContext (ctx);
 
+    gtk_widget_show_now (chooser);
+
+    /* can it ever happen that it'd be NULL here ??? */
+    if (chooser->window != NULL) {
+	    gdm_wm_init (GDK_WINDOW_XWINDOW (chooser->window));
+
+	    /* Run the focus, note that this will work no matter what
+	     * since gdm_wm_init will set the display to the gdk one
+	     * if it fails */
+	    gdm_wm_focus_window (GDK_WINDOW_XWINDOW (chooser->window));
+    }
+
     gtk_main();
 
     exit (EXIT_SUCCESS);
